/*
 * CLOTH - Command Line Options Templates Header
 * (c) 2002-2004 Rob Hague
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use, copy,
 * modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT.  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

#include <string>
#include <iostream>
#include <exception>
#include <map>
#include <assert.h>

using namespace std; // Required for MSVC, GCC 3 and maybe others

namespace cloth
{
  class OptI {
  public:
    OptI(char s, const char *l, const char* d)
      : sname(s), lname(l), desc(d) {};
    char sname;
    const char *lname;
    const char *desc;
    bool isImplied;
    virtual void setValue(char *)=0;
    virtual void setImplied() = 0;
  };

  class ParseError {
  public:
    static ParseError *current;
    ParseError(string s): message(s) { current = this; };
    string message;
  };

  class Parser {
    static Parser *instance;
    static Parser *theParser()
      { return instance?instance:(instance = new Parser()); };
    
    typedef map<char,OptI*> charmap;
    typedef map<string,OptI*> cstrmap;
    charmap sMap;
    cstrmap lMap;

    char *progname;

    class Lexer {
      int argc; char **argv, *point;
    public:
      Lexer(int c, char **v) : argc(c), argv(v), point(*v) {};

      char peekchar()
        { return *point; };
      char nextchar()
        { char c = *(point++); if (!*point) peekstring(); return c; };
      char *peekstring()
        { return point; };
      char *nextstring()
        {
          if (--argc < 0) throw ParseError("Expecting more arguments");
          char *s = point;
          point = *(++argv);
          return s;
        };
    };

  public:
    Parser(): progname(0) {};

    static void registerOpt(OptI* opt, char sname, const char *lname)
      {
        theParser()->sMap[sname] = opt; 
        if (lname) theParser()->lMap[lname] = opt; 
      };


    static void parse(int argc, char **argv)
      {
        try
          {
            parse_or_throw(argc,argv);
          }
        catch (ParseError e)
          {
            cout << e.message << endl;
            Parser::help();
            exit(1);
          }
      }

    /*
     * The parser; a hand-written state machine. Doing it like this
     * removes the neccesity for any external tools such as Yacc, and
     * the language is simple enough that it shouldn't be too
     * difficult to follow.
     */
    static void parse_or_throw(int argc, char **argv) throw(ParseError)
      {
        enum { START, ONEDASH, TWODASH, SHORT, END, VALUE } state=START;

        theParser()->progname = *(argv++); argc--;
        Lexer lex(argc,argv); char *longname, shortname; OptI *opt;

        while (state != END) switch (state)
          {
          case START:
            if (lex.peekstring())
              {
                if (lex.peekchar() == '-') { lex.nextchar(); state=ONEDASH; }
                else lex.nextstring();
              }
            else
              { state = END; };
            break;
          case ONEDASH:
            if (lex.peekchar() == '-') { lex.nextchar(); state=TWODASH; }
            else { state = SHORT; };
            break;
          case TWODASH:
            longname = lex.nextstring();
            opt = theParser()->lMap[longname];
            if (!opt) throw ParseError("No such option");
            if (opt->isImplied)
              {
                opt->setImplied();
                state = START;
              }
            else
              {
                state = VALUE;
              }
            break;
          case SHORT:
            shortname = lex.nextchar();
            opt = theParser()->sMap[shortname];
            if (!opt) throw ParseError("No such option");
            if (opt->isImplied)
              {
                opt->setImplied();
                state = lex.peekchar()?SHORT:START;
              }
            else
              {
                if (!lex.peekchar()) lex.nextstring();
                state = VALUE;
              };
            break;
          case VALUE:
            opt->setValue(lex.nextstring());
            state = START;
          };
      };

    static void help()
      {
        for(charmap::const_iterator i = theParser()->sMap.begin();
            i != theParser()->sMap.end(); i++)
          {
            OptI* opt = i->second;
            if (!opt) continue; // Failing to find something with [] inserts a
                                // mapping from it to NULL
            cout << "-" << opt->sname;
            int l = 0;
            if (opt->lname)
              {
                cout << ", --" << opt->lname;
                l = strlen(opt->lname)+4;
              }
            if (!opt->isImplied) { cout << " <val>"; l += 6; };
            while (l++ < 16) cout << ' ';
            cout << ' ' << opt->desc << "\n";
          }
      }
  };

  template<class T> class Opt: OptI
  {
    T value, implied;
    
  public:
    Opt<T>(char s, const char* l, const char* d,T def, T imp):
      value(def), implied(imp), OptI(s,l,d)
      { isImplied=true; Parser::registerOpt(this,s,l); };

    Opt<T>(char s, const char* l, const char* d,T def = T()):
      value(def), OptI(s,l,d)
      { isImplied=false; Parser::registerOpt(this,s,l); };
    
    operator const T() { return value; };

    virtual void setValue(char *v)
      {
        convert(value,v);
      }

    virtual void setImplied()
      {
        value = implied;
      }
  };

  // A utility class for boolean flags, and one for help arguments
  class Flag: public Opt<bool>
  {
  public:
    Flag(char s, const char* l, const char* d):
      Opt<bool>(s,l,d,false,true) {};
  };

  class Help: Flag
  {
  public:
    Help(char s, const char* l,
         const char* d="Display a usage message and exit."): Flag(s,l,d) {};
    virtual void setImplied()
      {
        Parser::help();
        exit(0);
      }
  };

  /*
   * Conversion functions
   *
   * The function void convert(T&,char*) is called by
   * Opt<T>::setValue, and hence there must exist an overridden
   * version for each type of value you use in an instantiated
   * option. The following are faily straightforward, but do not do
   * error detection. If you wish to define your own conversion
   * functions for the types here (int, bool, char* and string),
   * #define CLOTH_NO_CONVERSIONS to omit these declarations.
   */

#ifndef CLOTH_NO_CONVERSIONS
  void convert(int &i, char *v) { i = atoi(v); };
  void convert(bool &i, char *v) { i = v?true:false; };
  void convert(char *&i, char *v) { i = v; };
  void convert(string &i, char *v) { i = v; };
#endif

  // Better include some storage for static members...
  Parser *Parser::instance = 0;
  ParseError *ParseError::current = 0;
};
